import time
import threading
from logging import getLogger
logger = getLogger('pyscab.'+__name__)

def get_required_time(plans, data):
    end_times = list()
    for plan in plans:
        end_time = data.get_length_by_id(plan[1])
        end_time += plan[0]
        end_times.append(end_time)
    return max(end_times)

class StimulationController(object):

    def __init__(self,
                 AudioHardwareController,
                 marker_send,
                 correct_latency = True,
                 correct_hardware_buffer = False,
                 time_tick = 0.0001,
                 share=None):
        """
        class for playing stimulating plan.

        Attributes
        ----------
        AudioHardwareController : An instance of pyscab.AudioHardwareController class
        marker_send : reference to a function
            A reference to function for send a marker
        mode : str {'serial', 'pararell'} default = 'serial'
            If it's set to 'pararell', it can be controll from parent process.
        time_tick : float, default=0.0001
            pause between each loop in play() function. In terms of real time, it should be reduced.
            However, it also caused intence compulational load.
        share : list or instance of multiprocessing class
            0 : before playing
            1 : playing
            2 : finish playing

            share[0] is used for sharing runnin status.
            When it's working in pararell mode, it will be stop playing immediate after share is changed to other than 1.
            e.g., when you are presenting stimulus in online BCI application with dynamic stopping,
            you need to stop playing when it's triggered. In that case, you need to set mode to pararell and use share variable.

            share[1] is used for sharing current marker value.

        """
        self.ahc = AudioHardwareController
        self.time_tick = time_tick
        self.correct_latency = correct_latency
        if self.correct_latency:
            self.offset = self.ahc.frames_per_buffer/self.ahc.frame_rate
            if correct_hardware_buffer is not False:
                if correct_hardware_buffer is True:
                    self.offset += self.ahc.device['defaultLowOutputLatency']
                else:
                    self.offset += correct_hardware_buffer/self.ahc.frame_rate
            self.marker_send_hw = marker_send
            self.marker_send = self.marker_send_offset
        else:
            self.marker_send = marker_send
        if share is None:
            self.share = [0 for m in range(8)]
        else:
            self.share = share
        logger.debug("time_tick for Stimulation Controller was set to %s", str(self.time_tick))

    def marker_send_offset(self, val):
        t = threading.Thread(target=self.marker_send_offset_thread, args=(val,))
        t.start()

    def marker_send_offset_thread(self, val):
        time.sleep(self.offset)
        self.marker_send_hw(val=val)

    def play(self, plans, data, time_termination = 'auto', pause=0.5):

        # initialize
        self.share[0] = 0

        del_idxs = list()
        if time_termination is None:
            # TODO : prep two functions for main_loop with and w/o termination with time
            #        to avoid verbosed conditional branch in loop
            time_termination = float('inf')
        elif time_termination.lower() == 'auto':
            time_termination = get_required_time(plans, data)
        
        logger.debug("session time was set to %s." ,str(time_termination))

        self.share[0] = 1
        self.ahc.open()
        logger.debug("Audio Hardware Controller Opening.")

        # requires time to be opened. with out this line, time_info won't be get
        # TO DO : get the state of instance from pyaudio and wait until it's opened instead of waiting with sleep
        #time.sleep(1)
        while self.ahc.get_time_info() is None:
            time.sleep(0.01)

        start = self.ahc.get_time_info()['current_time']
        while self.share[0] == 1:
            now = self.ahc.get_time_info()['current_time'] - start
            for idx, plan in enumerate(plans):
                if now > plan[0]:
                    self.ahc.play(data.get_data_by_id(plan[1]),plan[2])
                    self.marker_send(val=plan[3])
                    self.share[1] = plan[3]
                    del_idxs.append(int(idx))
                    logger.debug("Playing, id:%s, ch:%s, marker:%s, path:%s",str(plan[1]),str(plan[2]),str(plan[3]),data.get_path_by_id(plan[1]))
                for del_idx in del_idxs:
                    del plans[del_idx]
                del_idxs=list()
            time.sleep(self.time_tick)
            if now > time_termination:
                self.share[0] = 2

        time.sleep(pause)
        self.ahc.close()
        logger.debug("Audio Hardware Controller Closing.")


import numpy as np
import time
import os
import wave
from logging import getLogger
logger = getLogger('pyscab.'+__name__)

class DataHandler(object):
    """
    Handling wav files and pcm data.

    Parameters
    ----------
    frame_rate : int, default=44100
        frame rate of data.
    verbose : bool, default=False
        verbosity while data loading.

    Attributes
    ----------
    pcm_data : list
        loaded audio data.
    n_ch : list of int
        number of channels of loaded audio data.
    n_frames : list of int
        number of frames of loaded audio data.
    paths : list of str
        path of loaded audio file.
    id : list of int
        id of loaded audio data.
    volume : list of float
        volume of loaded audio data.
    frame_rate : int
        frame rate of data
    sample width : list of int
        sample width in byte of loaded audio data.
    dtype : np.dtype
        data format type.
    """   

    def __init__(self, frame_rate = 44100, verbose=False):
        self.pcm_data = list()
        self.n_ch = list()
        self.n_frames = list()
        self.paths = list()
        self.id = list()
        self.volume = list()
        self.frame_rate = frame_rate
        self.sample_width = list()
        self.verbose = verbose
        self.dtype = np.dtype("int16")

    def load(self, id, path, volume=1.0):
        """
        Load wav file by path.

        Parameters
        ----------
        id : int
            id for this audio data.
        path : str
            file path of audio file to be loaded.
        volume : float, default=1.0
            volume for this audio data.

        Raises
        ------
        ValueError
            Raises ValueError if existing (already been used) id is passed to id.
        ValueError
            Raises ValueError if format of loaded wav file is not Int16 or UInt8.

        Examples
        --------
        You just need to specify the id for data, and file path to be loaded.

        >>> import pyscab
        >>> dh = pyscab.DataHandler()
        >>> dh.load(1, path = "/home/USER/Music/something.wav", volume = 0.5)
        """
        if self._is_exist_id(id) is False:
            self.id.append(id)
        else:
            raise ValueError("Passed id " + str(id) + " is dumplicated.")

        f_name = os.path.basename(path)

        wf = wave.open(path, 'rb')
        n_ch = wf.getnchannels()

        # it should be set in __init__()
        sw = wf.getsampwidth()
        if sw == 1:
            self.dtype = dtype = np.dtype("uint8")
        elif sw == 2:
            self.dtype = dtype = np.dtype("int16")
        else:
            raise ValueError("file must be Int16 or UInt8 PCM wave format")

        nf = wf.getnframes()

        if self.verbose:
            start = time.time()
            print("start loading : " + f_name)

        logger.debug("start loading : %s", path)
        data = np.zeros((nf, n_ch)).astype(dtype)

        data_wav = wf.readframes(nf)
        data_wav = np.frombuffer(data_wav, dtype=dtype)
        data_wav = data_wav * volume
        data_wav = data_wav.astype(dtype)
        for ch in range(n_ch):
            data[:, ch] = data_wav[ch::n_ch]

        if self.verbose:
            print("finish loading : " + str(f_name))
            end = time.time()
            print("elapsed time : " + str(end-start))

        self.paths.append(path)
        self.pcm_data.append(data)
        self.n_ch.append(n_ch)
        self.sample_width.append(sw)
        self.n_frames.append(nf)
        self.volume.append(volume)

    def apply_window(self, id, t_raise_fall):
        """
        Applying window function to data specified by id.

        Parameters
        ----------
        id : int
            data id to be applied filter function.
        t_raise_fall : float
            time duration for raise and fall.

        Notes
        -----
        If r_raise_fall is set to 0.05. Both raising and falling time will be set to 0.05.
        Currently, only linear function can be applied. Other function will be implemented in the future.
        """

        n_frames = self.get_nframes_by_id(id)
        fs = self.frame_rate

        raise_ = np.arange(0, int(t_raise_fall*fs))
        fall_ = np.arange(int(t_raise_fall*fs), 0, -1)

        raise_ = raise_ / max(raise_)
        fall_ = fall_ / max(fall_)

        flat = np.ones((n_frames - np.size(raise_) - np.size(raise_)))

        win_filter = np.concatenate([raise_, flat, fall_])

        idx = self._id2idx(id)
        n_ch = self.n_ch[idx]

        win = np.zeros((np.size(win_filter), n_ch))
        for m in range(n_ch):
            win[:, m] = win_filter

        self.pcm_data[idx] = self.pcm_data[idx]*win
        self.pcm_data[idx] = self.pcm_data[idx].astype(self.dtype)
    
    def add_pcm(self, id, data):
        """
        Add the pcm audio data (matrix) to DataHandler.

        Parameters
        ----------
        id : int
            id for this audio data.
        data : np.ndarray
            audio data. shoule have shape of (number of samples, number of channels)

        Notes
        -----
        File path will be automatically set to "PCM"

        Raises
        ------
        ValueError
            Raises ValueError if existing (already been used) id is passed to id.
        """
        if self._is_exist_id(id) is False:
            self.id.append(id)
        else:
            raise ValueError("Passed id " + str(id) + " is dumplicated.")

        if data.ndim == 1:
            data = np.atleast_2d(data)
            data = data.transpose()

        self.paths.append("PCM")
        self.pcm_data.append(data)
        self.n_frames.append(data.shape[0])
        self.n_ch.append(data.shape[1])

    def get_nframes_by_id(self, id):
        """
        get number of frames of the audio data specified by id.

        Parameters
        ----------
        id : int
            id of audio data.

        Returns
        -------
        number_of_frames : int
            number of frames of the audio data specified by id.
        """
        return np.shape(self.get_data_by_id(id))[0]

    def get_length_by_id(self, id):
        """
        get length (in seconds) of the audio data specified by id.

        Parameters
        ----------
        id : int
            id of audio data.

        Returns
        -------
        length : int
            length (in seconds) of the audio data specified by id.
        """
        # return type in second.
        return self.get_nframes_by_id(id)/self.frame_rate

    def get_data_by_id(self, id):
        """
        get data of the audio data specified by id.

        Parameters
        ----------
        id : int
            id of audio data.

        Returns
        -------
        data : np.ndarray
            audio data array specified by id.
        """
        idx = self._id2idx(id)
        return self.pcm_data[idx]

    def get_n_ch_by_id(self, id):
        """
        get number of channels of the audio data specified by id.

        Parameters
        ----------
        id : int
            id of audio data.

        Returns
        -------
        number_of_channels : int
            number of channels of audio data specified by id.
        """
        idx = self._id2idx(id)
        return self.n_ch[idx]

    def get_path_by_id(self,id):
        """
        get file path of the audio data specified by id.

        Parameters
        ----------
        id : int
            id of audio data.

        Returns
        -------
        path : str
            file path of audio data specified by id.
        """
        idx = self._id2idx(id)
        return self.paths[idx]

    def _id2idx(self, id):
        idx = np.where(np.array(self.id) == id)
        if len(idx[0]) == 0:
            return None
            #raise ValueError("id does not exist.")
        else:
            return int(idx[0])

    def _is_exist_id(self, id):
        if self._id2idx(id) == None:
            return False
        else:
            return True


import pyscab

file_dirs = list()
file_dirs.append("./tests/440Hz_stereo.wav")


#------------------------------------------------
# load audio files
#

afh = pyscab.DataHandler()
afh.load(1, file_dirs[0], volume = 0.7)

#------------------------------------------------
# generate sine wave
#

sine = pyscab.generate_pcm(number_of_channels=2, frequency=440, duration=1, format="INT16", volume=0.7, window=None)
afh.add_pcm(2, sine)

#------------------------------------------------
# audio play plan
#

# plan format
# [time, file_id, ch, marker]

audio_plan = list()

audio_plan.append([0.5, 1, [1,2], 1])
audio_plan.append([1.5, 2, [1,2], 2])


#------------------------------------------------
# open and initialize Audio I/F
#

DEVICE_NAME = 'X-AIR ASIO Driver'
#DEVICE_NAME = 'ASIO4ALL v2'
#DEVICE_NAME = 'default'
ahc = pyscab.AudioInterface(device_name = DEVICE_NAME, n_ch = 2, format="INT16", frames_per_buffer = 512)

def marker_send(val):
    print("marker sent : " + str(val))

from multiprocessing import Array
share = Array('d', range(8))
share = [0 for m in range(8)]

stc = pyscab.StimulationController(ahc, marker_send=marker_send, share=share, correct_latency = True, correct_hardware_buffer = True)
stc.play(audio_plan, afh)


import pyaudio
import numpy as np
from logging import getLogger
logger = getLogger('pyscab.'+__name__)

class CallbackParams(object):
    """
    class for referencing parameters of callback function of portaudio.

    Attributes
    ----------
    data : list of pyscab.ReadAudioChunk
        containing audio data to play.
    data_finished : list of bool
        True if all audio data in attribute data which has corresponding index were played.
    """
    def __init__(self):
        self.init()
    
    def init(self):
        """
        initialize all parameters hold by CallbackParams instance.
        """
        self.data = list()
        self.data_finished = list()
        self.format = None
        self.time = None
        self.n_ch = None
        self.data_callback = None

callback_params = CallbackParams()

def callback(in_data, frame_count, time_info, status, p=callback_params):
    #data = np.zeros((frame_count, p.n_ch), dtype=p.format)
    p.data_callback.fill(0)
    for idx, is_finished in enumerate(p.data_finished):
    #for idx_audio, audio in enumerate(p.data):
    #for audio in p.data:
        if is_finished is False:
            chunk = p.data[idx].read_chunk()
            for idx_ch, ch in enumerate(p.data[idx].get_ch()):
                p.data_callback[:, ch-1] += chunk[:, idx_ch]
        #else:
        #    p.data_finished[idx] = True
            # if all audio chunks were loaded, delete the instance
            #
            # deleted this in ver. 0.0.6, since doing this will take time and makes problem with presenting short audio with fast SOA.
            #del p.data[idx_audio]
    p.time = time_info
    return (np.ravel(p.data_callback, order='C'), pyaudio.paContinue)

def show_devices():
    """
    print available devices connected to the computer.
    """
    pya = pyaudio.PyAudio()
    hi = HardwareInformation(pya)
    devices = hi.devices
    #print(devices)
    for device in devices:
        print("%s : " %(device) + str(devices[device]))
        print("\n")
        #pass
        #print(device)
        #print("name : '" + str(devices[device]['name']) + "', maxOutputChannels : " + str(devices[device]['maxOutputChannels']))


class HardwareInformation(object):
    """
    get hardware information of audio devices.

    Attributes
    ----------
    devices : dict
        containes device name and maximum number of output channels of each device. 
        keys : 'name', 'maxOutputChannels'
    """
    def __init__(self, pya):
        self.devices = dict()
        n_devs = pya.get_device_count()
        for idx in range(n_devs):
            device = pya.get_device_info_by_index(idx)
            name = device['name']
            if device['maxOutputChannels'] > 0 and name not in self.devices:
                self.devices[name] = device

    def get_output_device_with_name(self, device_name):
        """
        get device information by name.

        Parameters
        ----------
        device_name : str
            name of the device.

        Returns
        -------
        device : pyaudio.PaDeviceInfo
        """
        return self.devices[device_name]

class AudioInterface(object):
    """
    Class for handling Audio Interface device.

    Attributes
    ----------
    device_name : str
        name of the device to be opened.
    format : str {'INT16', 'UINT8'}
        format to be used for device.
    frame_rate : int
        frame rate of device.
    frames_per_butter : int
        frames per buffer
    pya : instance of pyaudio.PyAudio()
        instance of PyAudio device.
    device_idx : int
        device index of opened device.
    num_cannels : int
        number of channels to be opened.
    n_ReadAudioChunk_obj : int
        total number of ReadAudioChunk instance.

    See Also
    --------
    Class ReadAudioChunk : used for reading audio data by chunk.
    """
    def __init__(self,
                 device_name=None,
                 n_ch = 2,
                 format="INT16",
                 frame_rate=44100,
                 frames_per_buffer=512,):
        """
        set audio device to be used and its settings

        Parameters
        ----------
        device_name : str, default=None
            device name to be opened.
            If it's None, 'default' will be used for Linux and 'ASIO4ALL' for Windows.
        n_ch : int, default=2
            number of channels to be opened
        format : {'INT16', 'UINT8'}, default='INT16'
            audio data fortmat to be played
        frame_rate : int, default=44100
            frame rate of audio device.
        frames_per_butter : int, default=512
            frames per butter of audio device. Which means chunk size of audio data read in callback function will be set to this value.
        """
        if device_name is None:
            import platform
            if platform.system() == "Linux":
                self.device_name = "default"
            elif platform.system() == "Windows":
                self.device_name = "ASIO4ALL v2"
            else:
                raise ValueError("Unknown Operating System")
        else:
            self.device_name = device_name 

        self.format = format
        if format.upper() == "INT16":
            self.format_pyaudio = pyaudio.paInt16
            self.format_np = np.dtype("int16")
        elif format.upper() == "UINT8":
            self.format_pyaudio = pyaudio.paUInt8
            self.format_np = np.dtype("uint8")
        else:
            raise ValueError("Unknown Format : " + self.format + ". It can only take INT16 or UINT8.")
        self.frame_rate = frame_rate
        self.frames_per_buffer = frames_per_buffer

        self.pya = pya = pyaudio.PyAudio()
        hardware_information = HardwareInformation(pya)

        available_devices = hardware_information.devices
        logger.debug("Audio Device '%s' was selected.", self.device_name)
        self.device = available_devices[self.device_name]
        self.device_idx = self.device['index']
        self.num_channels = n_ch

        self.n_ReadAudioChunk_obj = 0


    def open(self):
        """
        open audio device.
        """
        callback_params.init()
        self.n_ReadAudioChunk_obj = 0
        callback_params.n_ch = self.num_channels
        callback_params.format = self.format_np
        callback_params.data_callback = np.zeros((self.frames_per_buffer, self.num_channels), dtype=self.format_np)
        self.stream = self.pya.open(format=self.format_pyaudio,
                                    channels=self.num_channels,
                                    frames_per_buffer=self.frames_per_buffer,
                                    output_device_index=self.device_idx,
                                    rate=self.frame_rate,
                                    output=True,
                                    stream_callback=callback)

    def play(self, data, ch):
        """
        play audio data from specified channel.

        Parameters
        ----------
        data : np.ndarray
            audio data which have a shape of (number of samples, number of channels)
        ch : list of int
            channel number of audio device to be played (start from 1)

        Examples
        --------
        In this case, audio_data will be played from 1st channel of audio device.

        >>> play(audio_data, [1])
        """
        callback_params.data.append(ReadAudioChunk(data, self.frames_per_buffer, ch, format = self.format_np, idx_obj = self.n_ReadAudioChunk_obj))
        callback_params.data_finished.append(False)
        self.n_ReadAudioChunk_obj += 1 # should be add 1 after executing appending ReadAudioChunk obj.

    def get_time_info(self):
        """
        get time information from portaudio (PaStreamCallbackTimeInfo)

        Returns
        -------
        time_info : dict
        """
        return callback_params.time

    def close(self):
        """
        stop audio stream and close device.
        """
        self.stream.stop_stream()
        self.stream.close()

    def terminate(self):
        """
        terminate portaudio.
        """
        self.pya.terminate()

class ReadAudioChunk(object):
    """
    reading audio data by chunk.

    Attributes
    ----------
    data : np.ndarray
        audio data.
    n_ch_data : int
        number of channels of audio data
    n_frames : int
        number of frames of audio data
    chunk_sixe : int
        chunk size.
    ch : list of int
        channel number of device which audio data will be played.
    format : np.dtype
        audio data format
    finished : Bool
        If True, all audio data have already been read. If False, there's remained data.
    idx_obj : int
        id of instance.
    """
    def __init__(self, data, chunk_size, ch, volume=1.0, format = np.dtype("int16"), idx_obj = None):
        """
        load data with specified settings

        Parameteres
        ----------
        data : np.ndarray
            audio data.
        chunk_sixe : int
            chunk size.
        ch : list of int
            channel number of device which audio data will be played.
        format : np.dtype, default=np.dtype('int16')
            audio data format
        idx_obj : int, default=None
            id of instance.
        """
        self.data = data.astype(format)
        self.n_ch_data = data.shape[1] # n_ch of data
        self.n_frames = data.shape[0]
        self.current_idx = 0
        self.chunk_size = chunk_size
        self.ch = ch # ch num to be stimuli presented
        self.format = format
        #self.remained_frames = self.n_frames
        self.finished = False
        self.idx_obj = idx_obj # will be change to id
        self.chunk_data = np.zeros((self.chunk_size, self.n_ch_data), dtype=self.format)
        # if mono audio will be played from multiple channel,
        # concatenate it to multiple channel matrix
        # in order to make computational load in portaudio callback function less
        while self.n_ch_data < len(self.ch):
            self.data = np.hstack((self.data,self.data))
            self.n_ch_data = self.data.shape[1]

    def read_chunk(self):
        """
        read audio data by chunk.

        Returns
        -------
        chunk_data : np.ndarray
            chunk data which have a shape of (chunk size, number of channels)
        """

        start = self.current_idx*self.chunk_size
        end = (self.current_idx*self.chunk_size)+self.chunk_size

        if end >= (self.n_frames-1):
            self.chunk_data.fill(0)
            self.chunk_data[0:(self.n_frames-start-1),:] = self.data[start:-1,:]
            
            #=====================
            #This is old method.
            #
            #deleted in ver 0.0.6 since it is slow.
            #=====================
            # padding zero to last block 
            #data = self.data[start:-1,:]
            #n_padding = self.chunk_size - data.shape[0]
            #data = np.vstack((data, np.zeros((n_padding, self.n_ch_data), dtype=self.format)))
            #data = np.vstack((data, np.zeros((n_padding, self.n_ch_data))))
            #self.finished = True
            #=====================
            callback_params.data_finished[self.idx_obj] = True
        else:
            self.chunk_data = self.data[start:end,:]
            #self.remained_frames -= self.chunk_size
        self.current_idx += 1
        return self.chunk_data

    def is_finished(self):
        """
        check if it's finished

        Returns
        -------
        is_finished : bool
            True if all data already been read, False if there's remained data.
        """
        return self.finished

    def get_ch(self):
        """
        get channel number to play the audio data.

        Returns
        -------
        channel : list of int
            channel number which audio data will be played.
        """
        return self.ch
